#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifndef N
#define N 10000
#endif

#ifndef CACHE_LINE_SIZE
#define CACHE_LINE_SIZE 64
#endif

double a[N][N] __attribute__ ((aligned(CACHE_LINE_SIZE)));

double wtime()
{
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return ts.tv_sec + ts.tv_nsec * 1E-9;
}

void matrix_init(double a[N][N])
{
    srand(0);
    for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
            a[i][j] = rand() % 100;
        }
    }
}

double matrix_sum(double a[N][N])
{
    double sum = 0;
    for (int j = 0; j < N; j++) {
        for (int i = 0; i < N; i++) {
            sum += a[i][j];  /* stride-N read */
        }
        /*
         * Inner loop: N cache misses (uses N new cache lines)
         * May evicts previous lines from the cache
         * Example: L1d 48KB 12 way, 64 sets = 768 lines
         *          if i > 768 cache starts evict prev lines
         */
    }
    return sum;
}

double matrix_sum_interchanged(double a[N][N])
{
    double sum = 0;
    for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
            sum += a[i][j];  /* stride-1 read */
        }
        /* Inner loop: uses N / 8 cache lines */
    }
    return sum;
}

int main()
{
    int NREPS = 10;

    double t = 0;
    for (int i = 0; i < NREPS; i++) {
        matrix_init(a);
        t -= wtime();
        matrix_sum(a);
        //matrix_sum_interchanged(a);
        t += wtime();
    }
    t /= NREPS;
    printf("# Matrix sum: N=%d, elapsed time (sec) %.6f, Melems/sec %.1f\n", N, t, N * N * 1E-6 / t);

    return 0;
}

